﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO.Ports;
using System.Drawing.Drawing2D;

namespace povGlobeControl
{
    public partial class Form1 : Form
    {
        PovIO povglobe;
        Bitmap bmpOriginal, bmpConverted, bmpForDevice;
        uint[][] frameConv = new uint[80][]; //80 columns by 4 uints per col
        uint[][] frameForDevice = new uint[80][]; //80 columns by 4 uints per col

        public Form1()
        {
            InitializeComponent();
            for (int i = 0; i < frameConv.Length; i++)
            {
                frameConv[i] = new uint[4];
            }
            for (int i = 0; i < frameForDevice.Length; i++)
            {
                frameForDevice[i] = new uint[4];
            }
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            cmbPortNo.Items.AddRange(SerialPort.GetPortNames());
            cmbPortNo.Text = cmbPortNo.Items[0].ToString();

            bmpOriginal = (Bitmap)picRealPrev.Image;
        }

        private void btnSerialConnect_Click(object sender, EventArgs e)
        {
            try
            {
                lblSerialLive.Text = "Connecting...";
                lblSerialLive.ForeColor = Color.Blue;
                lblSerialLive.Visible = true;
                povglobe = new PovIO();
                povglobe.PortName = cmbPortNo.Text;
                povglobe.Connect();
                if (!povglobe.Ping())
                    throw new TimeoutException("No ping received from POV globe. Please try again.");
                cmbPortNo.Enabled = false;
                lblSerialLive.Text = "Live";
                btnSerialConnect.Enabled = false;
                btnSerialDisconnect.Enabled = true;
                gbxPovControl.Enabled = true;
                gbxImgSend.Enabled = true;
                timer1.Enabled = true; //successfully connected, update fps value
            }
            catch (Exception ex)
            {
                timer1.Enabled = false; //stop polling fps
                povglobe.Disconnect();
                lblSerialLive.Text = "Disconnected";
                lblSerialLive.ForeColor = Color.Black;
                lblSerialLive.Visible = false;
                MessageBox.Show(ex.Message, "Error!");
            }
        }

        private void btnSerialDisconnect_Click(object sender, EventArgs e)
        {
            if (povglobe == null)
                return;
            if (!povglobe.IsConnected)
                return;
            try
            {
                povglobe.Disconnect();
                cmbPortNo.Enabled = true;
                lblSerialLive.Text = "Disconnected";
                lblSerialLive.ForeColor = Color.Black;
                btnSerialConnect.Enabled = true;
                btnSerialDisconnect.Enabled = false;
                gbxPovControl.Enabled = false;
                gbxImgSend.Enabled = false;

            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Error!");
            }
        }

        private void btnGenHex_Click(object sender, EventArgs e)
        {
            try
            {
                //generate hex
                HexGen hgen = new HexGen(GenerateHex((Bitmap)picConverted.Image));
                hgen.ShowDialog();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Error!");
            }
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            if(povglobe == null)
                return;
            if(!povglobe.IsConnected)
                return;
            try
            {
                //update fps info
                int ticks = povglobe.GetTicks(); //last revolution count
                ticks = ticks * 1024; //respect timer prescaler to get real MCLK ticks
                double seconds = ticks / 32000000.0;
                //32Mhz clock, timer prescaler 1024
                double fps = 1.0 / seconds;
                lblFPS.Text = String.Format("{0:0.0}", fps);
            }
            catch (Exception)
            {
                //do nothing, skip and try again later
            }
        }

        private void rdScrollCW_CheckedChanged(object sender, EventArgs e)
        {
            if (povglobe == null)
                return;
            //send mode scroll CW and uncheck other 2
            try
            {
                povglobe.ScrollCW();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Error!");
            }
        }

        private void rdScrollOff_CheckedChanged(object sender, EventArgs e)
        {
            if (povglobe == null)
                return;
            //send mode scroll off and uncheck other 2
            try
            {
                povglobe.ScrollOff();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Error!");
            }
        }

        private void rdScrollCCW_CheckedChanged(object sender, EventArgs e)
        {
            if (povglobe == null)
                return;
            //send mode scroll ccw and uncheck other 2
            try
            {
                povglobe.ScrollCCW();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Error!");
            }

        }

        private void btnImgWrite_Click(object sender, EventArgs e)
        {
            if (!(povglobe != null && povglobe.IsConnected))
                return;
            if (!(frameForDevice != null && frameForDevice.Length == 80))
                return; //frame invalid
            try
            {
                povglobe.SetColumnAddress(0); //reset RAM address to be safe
                for (int x = 0; x < 80; x++)
                {
                    povglobe.WriteData(frameForDevice[x][0],
                        frameForDevice[x][1], frameForDevice[x][2], frameForDevice[x][3]);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Error!");
            }
        }

        private void btnOpenImg_Click(object sender, EventArgs e)
        {
            Bitmap temp = new Bitmap(textBox1.Text);
            bmpOriginal = new Bitmap(temp,new Size(80,40)); //downsize to correct pixels
            picRealPrev.Image = bmpOriginal;
        }

        public void ConvertBitmap(Bitmap bmp, int Rcutoff, int Gcutoff, int Bcutoff)
        {
            if (!(bmp != null && bmp.Height == 40 && bmp.Width == 80))
                throw new ArgumentNullException("Convert bitmap: check source bitmap.");
            uint[][] frame = new uint[80][];
            Bitmap newBmp = new Bitmap(80, 40);
            for (int x = 0; x < 80; x++)
            {
                uint d3 = 0, d2 = 0, d1 = 0, d0 = 0; //need these for each column
                for (int y = 0; y < 40; y++)
                {
                    Color cl = bmp.GetPixel(x, (39) - y);
                    int g = 0;
                    if (cl.G >= Gcutoff) g = 1;
                    int r = 0;
                    if (cl.R >= Rcutoff) r = 1;
                    int b = 0;
                    if (cl.B >= Bcutoff) b = 1;
                    newBmp.SetPixel(x, (39) - y, Color.FromArgb(255 * r, 255 * g, 255 * b));
                    int grb = g << 10 | r << 5 | b;
                    int mod8 = y % 8;
                    int mod2 = mod8 % 2;
                    int shiftInReg = (y / 8);
                    grb = grb << shiftInReg;
                    grb = grb << (mod2 * 15);
                    if (mod8 == 0 || mod8 == 1)
                    {
                        d0 = d0 | (uint)grb;
                    }
                    else if (mod8 == 2 || mod8 == 3)
                    {
                        d1 = d1 | (uint)grb;
                    }
                    else if (mod8 == 4 || mod8 == 5)
                    {
                        d2 = d2 | (uint)grb;
                    }
                    else if (mod8 == 6 || mod8 == 7)
                    {
                        d3 = d3 | (uint)grb;
                    }
                    frame[x] = new uint[4];
                    frame[x][0] = d3;
                    frame[x][1] = d2;
                    frame[x][2] = d1;
                    frame[x][3] = d0;
                }
                frameConv = frame;
                bmpConverted = newBmp;
            }
        }

        public string GenerateHex(Bitmap bmp)
        {
            if (!(bmp != null && bmp.Width == 80 && bmp.Height == 40))
                throw new ArgumentNullException("Generate Hex: Invalid bitmap.");
            StringBuilder sb = new StringBuilder();
            for (int x = 0; x < bmp.Width; x++)
            {
                StringBuilder lb = new StringBuilder();
                uint d3 = 0, d2 = 0, d1 = 0, d0 = 0; //need these for each column
                for (int y = 0; y < bmp.Height; y++)
                {
                    Color cl = bmp.GetPixel(x, (bmp.Height - 1) - y);
                    int g = cl.G >> 7; //only msb of colour is useful -> used if val >= 128
                    int r = cl.R >> 7;
                    int b = cl.B >> 7;
                    double lumi = 0.21 * r + 0.72 * b + 0.07 * g; //add a little ASCII rep of img
                    char cc = ' '; //to help orientation in hex file
                    if (lumi < 0.001) { }
                    else if (lumi < 0.25) cc = '/';
                    else if (lumi < 0.5) cc = '0';
                    else if (lumi < 0.75) cc = (char)244; //¶
                    else cc = (char)219; //█
                    lb.Append(cc);
                    int grb = g << 10 | r << 5 | b;
                    int mod8 = y % 8;
                    int mod2 = mod8 % 2;
                    int shiftInReg = (y / 8);
                    grb = grb << shiftInReg;
                    grb = grb << (mod2 * 15);
                    if (mod8 == 0 || mod8 == 1)
                    {
                        d0 = d0 | (uint)grb;
                    }
                    else if (mod8 == 2 || mod8 == 3)
                    {
                        d1 = d1 | (uint)grb;
                    }
                    else if (mod8 == 4 || mod8 == 5)
                    {
                        d2 = d2 | (uint)grb;
                    }
                    else if (mod8 == 6 || mod8 == 7)
                    {
                        d3 = d3 | (uint)grb;
                    }
                }
                string line = "0x" + d3.ToString("X8")
                    + ", 0x" + d2.ToString("X8")
                    + ", 0x" + d1.ToString("X8")
                    + ", 0x" + d0.ToString("X8") + ", ";
                sb.Append(line);
                sb.AppendLine("\t// " + lb.ToString());

            }
            return sb.ToString();
        }

        private void btnConvert_Click(object sender, EventArgs e)
        {
            try
            {
                ConvertBitmap(bmpOriginal, tbCutoffR.Value, tbCutoffG.Value, tbCutoffB.Value);
                picConverted.Image = bmpConverted;
                picEditZoom.Image = ZoomImg(bmpConverted);
                btnGenHex.Enabled = true; //from this point on, no need to disable as always some image will be laoded
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Error!");
            }

        }

        private Bitmap ZoomImg(Bitmap bmp)
        {
            //little quick and dirty helper to zoom with nearest neighbor instead of interpolation
            Bitmap retbmp = new Bitmap(320, 160);
            using (var g = Graphics.FromImage(retbmp))
            {
                g.FillRectangle(Brushes.Black, 0, 0, 320, 160);
                g.InterpolationMode = InterpolationMode.NearestNeighbor;
                g.DrawImage(bmp, new Rectangle(new Point(1,1), retbmp.Size));
            }
            return retbmp;
        }

        private void btnImgLoad_Click(object sender, EventArgs e)
        {
            if (bmpConverted == null)
                return;
            bmpForDevice = bmpConverted;
            frameForDevice = frameConv;
            picDevicePreview.Image = bmpForDevice;
        }

        private void tbCutoffR_Scroll(object sender, EventArgs e)
        {
            lblCutoffR.Text = tbCutoffR.Value.ToString();
        }

        private void tbCutoffG_Scroll(object sender, EventArgs e)
        {
            lblCutoffG.Text = tbCutoffG.Value.ToString();
        }

        private void tbCutoffB_Scroll(object sender, EventArgs e)
        {
            lblCutoffB.Text = tbCutoffB.Value.ToString();
        }

        private void btnGenHex_Click_1(object sender, EventArgs e)
        {
            if (bmpConverted == null)
                return;
            HexGen hg = new HexGen(GenerateHex(bmpConverted));
            hg.ShowDialog();
        }
    }
}
