﻿using System.Collections.Generic;
using System.Text;
using System.IO.Ports;
using System;

namespace povGlobeControl
{
    class PovIO
    {
        public enum Commands
        {
            Ping=0x1,
            Ticks,
            SetAddress,
            ReadData,
            WriteData,
            ScrollCW,
            ScrollCCW,
            ScrollOff
        }

        public enum Responses
        {
            ACK = 0x55,
            Error = 0xDB,
            Timeout = 0xFE
        }

        private SerialPort sp;

        public string PortName
        {
            get { return sp.PortName; }
            set { sp.PortName = value; }
        }

        public bool IsConnected
        {
            get { return sp.IsOpen; }
        }

        public PovIO()
        {
            sp = new SerialPort();
            sp.BaudRate = 115200;
            sp.ReadTimeout = 2000; //millis
            sp.WriteTimeout = 2000;
        }

        public void Connect()
        {
            sp.Open();
        }

        public void Disconnect()
        {
            sp.Close();
        }

        public bool Ping()
        {
            //if (sp.BytesToRead != 0)
            //   throw new InvalidOperationException("Receive buffer not clear, bad state.");
            byte[] cmd = {(byte)Commands.Ping};
            sp.Write(cmd,0,1);
            bool ret = false;
            try
            {
                ret = sp.ReadByte() == (int)Responses.ACK;
            }
            catch (TimeoutException)
            {
                //if timeout, ping is false, no other consequences
            }
            return ret;
        }

        public int GetTicks()
        {
            //return ticks of last revolution, use to calculate FPS
            int ret = 0;
            byte[] cmd = { (byte)Commands.Ticks };
            sp.Write(cmd, 0, 1);
            if (sp.ReadByte() != (int)Responses.ACK)
            {
                throw new InvalidOperationException("Serial Error, no valid answer received");
            }
            int highbyte = sp.ReadByte();
            int lowbyte = sp.ReadByte();
            ret = highbyte << 8 | lowbyte; //put together 16bit counter val from both bytes

            return ret;
        }

        public int SetColumnAddress(int col)
        {
            int ret = 0;
            byte[] cmd = { (byte)Commands.SetAddress};
            sp.Write(cmd, 0, 1);
            if (sp.ReadByte() != (int)Responses.ACK)
            {
                throw new InvalidOperationException("Serial Error, no valid answer received");
            }
            ret = sp.ReadByte();

            return ret;
        }

        public int[] ReadData()
        {
            int[] ret = new int[] { 0, 0, 0, 0 };
            byte[] cmd = { (byte)Commands.ReadData };
            sp.Write(cmd, 0, 1);
            if (sp.ReadByte() != (int)Responses.ACK)
            {
                throw new InvalidOperationException("Serial Error, no valid answer received");
            } //valid ACK, next 16 bytes should be data (or timeout will occur)
            ret[3] = sp.ReadByte() << 24 | sp.ReadByte() << 16 | sp.ReadByte() << 8 | sp.ReadByte();
            ret[2] = sp.ReadByte() << 24 | sp.ReadByte() << 16 | sp.ReadByte() << 8 | sp.ReadByte();
            ret[1] = sp.ReadByte() << 24 | sp.ReadByte() << 16 | sp.ReadByte() << 8 | sp.ReadByte();
            ret[0] = sp.ReadByte() << 24 | sp.ReadByte() << 16 | sp.ReadByte() << 8 | sp.ReadByte();

            return ret;
        }

        public void WriteData(uint d3, uint d2, uint d1, uint d0)
        {
            byte[] cmd = new byte[17];
            cmd[0] = (byte)(byte)Commands.WriteData;
            cmd[1] = (byte)(d3 >> 24); //highest byte
            cmd[2] = (byte)(d3 >> 16);
            cmd[3] = (byte)(d3 >> 8);
            cmd[4] = (byte)(d3);
            cmd[5] = (byte)(d2 >> 24);
            cmd[6] = (byte)(d2 >> 16);
            cmd[7] = (byte)(d2 >> 8);
            cmd[8] = (byte)(d2);
            cmd[9] = (byte)(d1 >> 24);
            cmd[10] = (byte)(d1 >> 16);
            cmd[11] = (byte)(d1 >> 8);
            cmd[12] = (byte)(d1);
            cmd[13] = (byte)(d0 >> 24);
            cmd[14] = (byte)(d0 >> 16);
            cmd[15] = (byte)(d0 >> 8);
            cmd[16] = (byte)(d0);
            sp.Write(cmd, 0, 17);
            if (sp.ReadByte() != (int)Responses.ACK)
            {
                throw new InvalidOperationException("Serial Error, no valid answer received");
            }
        }

        public void ScrollCW()
        {
            byte[] cmd = { (byte)Commands.ScrollCW };
            sp.Write(cmd, 0, 1);
            if (sp.ReadByte() != (int)Responses.ACK)
            {
                throw new InvalidOperationException("Serial Error, no valid answer received");
            }

        }

        public void ScrollCCW()
        {
            byte[] cmd = { (byte)Commands.ScrollCCW };
            sp.Write(cmd, 0, 1);
            if (sp.ReadByte() != (int)Responses.ACK)
            {
                throw new InvalidOperationException("Serial Error, no valid answer received");
            }

        }

        public void ScrollOff()
        {
            byte[] cmd = { (byte)Commands.ScrollOff };
            sp.Write(cmd, 0, 1);
            if (sp.ReadByte() != (int)Responses.ACK)
            {
                throw new InvalidOperationException("Serial Error, no valid answer received");
            }

        }
    }
}
